@@ -31,12 +31,16 @@ from .tools import (
     log_norm,
     select_features,
     pca,
-    pc_regress_out,
+    pc_transform,
+    tsvd,
+    tsvd_transform,
+    regress_out,
     highly_variable_features,
     set_group_attribute,
     correct_batch,
     run_harmony,
     run_scanorama,
+    get_neighbors,
     neighbors,
     calc_kBET,
     calc_kSIM,
@@ -49,10 +53,8 @@ from .tools import (
     spectral_louvain,
     spectral_leiden,
     tsne,
-    fitsne,
     umap,
     fle,
-    net_tsne,
     net_umap,
     net_fle,
     de_analysis,
@@ -61,7 +63,6 @@ from .tools import (
     find_markers,
     infer_path,
     calc_signature_score,
-    run_scrublet,
     infer_doublets,
     mark_doublets,
 )
@@ -82,7 +83,10 @@ from .plotting import (
     ridgeplot,
 )
 
-from importlib_metadata import version, PackageNotFoundError
+try:
+    from importlib.metadata import version, PackageNotFoundError
+except ImportError:  # < Python 3.8: Use backport module
+    from importlib_metadata import version, PackageNotFoundError
 
 try:
     __version__ = version('pegasuspy')
