@@ -18,8 +18,8 @@ from typing import List, Tuple, Union, Optional, Callable
 import logging
 logger = logging.getLogger(__name__)
 
-from pegasus.tools import X_from_rep
-from .plot_utils import _transform_basis, _get_nrows_and_ncols, _get_marker_size, _get_dot_size, _get_subplot_layouts, _get_legend_ncol, _get_palette, _plot_cluster_labels_in_heatmap, RestrictionParser, DictWithDefault, _generate_categories, _plot_corners
+from pegasus.tools import X_from_rep, slicing
+from .plot_utils import _transform_basis, _get_nrows_and_ncols, _get_marker_size, _get_dot_size, _get_subplot_layouts, _get_legend_ncol, _get_palette, RestrictionParser, DictWithDefault, _generate_categories, _plot_corners
 
 
 def scatter(
@@ -111,11 +111,11 @@ def scatter(
         attrs = [attrs]
     nattrs = len(attrs)
 
-    if isinstance(data, MultimodalData) or isinstance(data, UnimodalData):
+    if not isinstance(data, anndata.AnnData):
         cur_matkey = data.current_matrix()
 
     if matkey is not None:
-        assert isinstance(data, MultimodalData) or isinstance(data, UnimodalData)
+        assert not isinstance(data, anndata.AnnData)
         data.select_matrix(matkey)
 
     x = data.obsm[f"X_{basis}"][:, 0]
@@ -161,7 +161,7 @@ def scatter(
                         loc = data.var_names.get_loc(attr)
                     except KeyError:
                         raise KeyError(f"{attr} is neither in data.obs nor data.var_names!")
-                    values = data.X[:, loc].toarray().ravel() if issparse(data.X) else data.X[:, loc]
+                    values = slicing(data.X, col = loc)
 
                 selected = restr_obj.get_satisfied(data, attr)
 
@@ -251,7 +251,7 @@ def scatter(
                 ax.set_ylabel(f"{basis}2")
 
     # Reset current matrix if needed.
-    if (not isinstance(data, anndata.AnnData)):
+    if not isinstance(data, anndata.AnnData):
         if cur_matkey != data.current_matrix():
             data.select_matrix(cur_matkey)
 
@@ -349,10 +349,10 @@ def scatter_groups(
     >>> pg.scatter_groups(data, attr='louvain_labels', groupby='Individual', basis='tsne', nrows = 2, ncols = 4, alpha = 0.5)
     >>> pg.scatter_groups(data, attr='anno', groupby='Channel', basis='umap', categories=['new_cat1:channel1,channel2', 'new_cat2:channel3'])
     """
-    if isinstance(data, MultimodalData) or isinstance(data, UnimodalData):
+    if not isinstance(data, anndata.AnnData):
         cur_matkey = data.current_matrix()
     if matkey is not None:
-        assert isinstance(data, MultimodalData) or isinstance(data, UnimodalData)
+        assert not isinstance(data, anndata.AnnData)
         data.select_matrix(matkey)
 
     x = data.obsm[f"X_{basis}"][:, 0]
@@ -371,7 +371,7 @@ def scatter_groups(
             loc = data.var_names.get_loc(attr)
         except KeyError:
             raise KeyError(f"{attr} is neither in data.obs nor data.var_names!")
-        values = data.X[:, loc].toarray().ravel() if issparse(data.X) else data.X[:, loc]
+        values = slicing(data.X, col = loc)
 
     is_cat = is_categorical_dtype(values)
     if (not is_cat) and (not is_numeric_dtype(values)):
@@ -671,10 +671,10 @@ def violin(
     if not is_list_like(attrs):
         attrs = [attrs]
 
-    if isinstance(data, MultimodalData) or isinstance(data, UnimodalData):
+    if not isinstance(data, anndata.AnnData):
         cur_matkey = data.current_matrix()
     if matkey is not None:
-        assert isinstance(data, MultimodalData) or isinstance(data, UnimodalData)
+        assert not isinstance(data, anndata.AnnData)
         data.select_matrix(matkey)
 
     nrows = len(attrs)
@@ -689,7 +689,7 @@ def violin(
             obs_keys.append(key)
         else:
             if key not in data.var_names:
-                logger.warning("Cannot find gene {key}. Please make sure all genes are included in data.var_names before running this function!")
+                logger.warning(f"Cannot find gene {key}. Please make sure all genes are included in data.var_names before running this function!")
                 return None
             genes.append(key)
 
@@ -700,7 +700,7 @@ def violin(
     if len(obs_keys) > 0:
         df_list.append(data.obs[obs_keys].reset_index(drop=True))
     if len(genes) > 0:
-        expr_mat = data[:, genes].X.toarray()
+        expr_mat = slicing(data[:, genes].X)
         df_list.append(pd.DataFrame(data=expr_mat, columns=genes))
     df = pd.concat(df_list, axis = 1)
 
@@ -738,13 +738,18 @@ def violin(
 
 def heatmap(
     data: Union[MultimodalData, UnimodalData, anndata.AnnData],
-    genes: Union[str, List[str]],
+    attrs: Union[str, List[str]],
     groupby: str,
     matkey: Optional[str] = None,
     on_average: bool = True,
     switch_axes: bool = False,
-    row_cluster: Optional[bool] = None,
-    col_cluster: Optional[bool] = None,
+    attrs_cluster: Optional[bool] = False,
+    attrs_dendrogram: Optional[bool] = True,
+    groupby_cluster: Optional[bool] = True,
+    groupby_dendrogram: Optional[bool] = True,
+    attrs_labelsize: Optional[float] = 10.0,
+    groupby_labelsize: Optional[float] = 10.0,
+    cbar_labelsize: Optional[float] = 10.0,
     panel_size: Tuple[float, float] = (10, 10),
     return_fig: Optional[bool] = False,
     dpi: Optional[float] = 300.0,
@@ -758,21 +763,35 @@ def heatmap(
 
     data: ``AnnData`` or ``MultimodalData`` or ``UnimodalData`` object
         Single-cell expression data.
-    genes: ``str`` or ``List[str]``
-        Features to plot.
+    attrs: ``str`` or ``List[str]``
+        Cell attributes or features to plot.
+        Cell attributes must exist in ``data.obs`` and must be numeric.
+        Features must exist in ``data.var``.
+        By default, attrs are plotted as columns.
     groupby: ``str``
         A categorical variable in data.obs that is used to categorize the cells, e.g. Clusters.
+        By default, data.obs['groupby'] is plotted as rows.
     matkey: ``str``, optional, default: ``None``
         If matkey is set, select matrix with matkey as keyword in the current modality. Only works for MultimodalData or UnimodalData objects.
     on_average: ``bool``, optional, default: ``True``
         If ``True``, plot cluster average gene expression (i.e. show a Matrixplot); otherwise, plot a general heatmap.
     switch_axes: ``bool``, optional, default: ``False``
-        By default, X axis is for genes, and Y axis for clusters. If this parameter is ``True``, switch the axes.
+        By default, X axis is for attributes, and Y axis for clusters. If this parameter is ``True``, switch the axes.
         Moreover, with ``on_average`` being ``False``, if ``switch_axes`` is ``False``, ``row_cluster`` is enforced to be ``False``; if ``switch_axes`` is ``True``, ``col_cluster`` is enforced to be ``False``.
-    row_cluster: ``bool``, optional, default: ``False``
-        Cluster rows and generate a row-wise dendrogram.
-    col_cluster: ``bool``, optional, default: ``True``
-        Cluster columns and generate a column-wise dendrogram.
+    attrs_cluster: ``bool``, optional, default: ``False``
+        Cluster attributes and generate a attribute-wise dendrogram.
+    attrs_dendrogram: ``bool``, optional, default: ``True``
+        Only matters if attrs_cluster is True. Show the dendrogram if this option is True.
+    groupby_cluster: ``bool``, optional, default: ``True``
+        Cluster data.obs['groupby'] and generate a cluster-wise dendrogram.
+    groupby_dendrogram: ``bool``, optional, default: ``True``
+        Only matters if groupby_cluster is True. Show the dendrogram if this option is True.
+    attrs_labelsize: ``float``, optional, default: 10.0
+        Fontsize for labels of attrs.
+    groupby_labelsize: ``float``, optional, default: 10.0
+        Fontsize for labels of data.obs['groupby'].
+    cbar_labelsize: ``float``, optional, default: 10.0
+        Fontsize of the color bar.
     panel_size: ``Tuple[float, float]``, optional, default: ``(10, 10)``
         Overall size of the heatmap in ``(width, height)`` form.
     return_fig: ``bool``, optional, default: ``False``
@@ -794,20 +813,41 @@ def heatmap(
     >>> pg.heatmap(data, genes=['CD14', 'TRAC', 'CD34'], groupby='louvain_labels')
 
     """
-    if isinstance(data, MultimodalData) or isinstance(data, UnimodalData):
+    if not isinstance(data, anndata.AnnData):
         cur_matkey = data.current_matrix()
     if matkey is not None:
-        assert isinstance(data, MultimodalData) or isinstance(data, UnimodalData)
+        assert not isinstance(data, anndata.AnnData)
         data.select_matrix(matkey)
 
-    if row_cluster is None:
-        row_cluster = True if switch_axes else False
+    if isinstance(attrs, str):
+        attrs = [attrs]
 
-    if col_cluster is None:
-        col_cluster = True if not switch_axes else False
+    obs_keys = []
+    genes = []
+    for key in attrs:
+        if key in data.obs:
+            assert is_numeric_dtype(data.obs[key])
+            obs_keys.append(key)
+        else:
+            if key not in data.var_names:
+                logger.warning(f"Cannot find gene {key}. Please make sure all genes are included in data.var_names before running this function!")
+                return None
+            genes.append(key)
 
-    df = pd.DataFrame(data[:, genes].X.toarray(), index=data.obs.index, columns=genes)
-    df['cluster_name'] = data.obs[groupby]
+    clusters = data.obs[groupby].values
+    if not is_categorical_dtype(clusters):
+        clusters = pd.Categorical(clusters)
+    else:
+        clusters.remove_unused_categories(inplace = True)
+    df_list = [pd.DataFrame({'cluster_name': clusters})]
+
+    if len(obs_keys) > 0:
+        df_list.append(data.obs[obs_keys].reset_index(drop=True))
+    if len(genes) > 0:
+        expr_mat = slicing(data[:, genes].X)
+        df_list.append(pd.DataFrame(data=expr_mat, columns=genes))
+    df = pd.concat(df_list, axis = 1)
+    attr_names = df.columns[1:].values
 
     if on_average:
         if not 'cmap' in kwargs.keys():
@@ -815,78 +855,107 @@ def heatmap(
         df = df.groupby('cluster_name').mean()
         cluster_ids = df.index
     else:
-        row_cluster = False if not switch_axes else row_cluster
-        col_cluster = False if switch_axes else col_cluster
-
-        cluster_ids = pd.Categorical(data.obs[groupby])
-        idx = cluster_ids.argsort()
-        df = df.iloc[idx, :]  # organize df by category order
-        df.drop(columns=['cluster_name'], inplace=True)
+        cluster_ids = df.pop('cluster_name').values
+        if not groupby_cluster:
+            idx = cluster_ids.argsort(kind = 'mergesort')
+            df = df.iloc[idx, :]  # organize df by category order
+            cluster_ids = cluster_ids[idx]
 
         cell_colors = np.zeros(df.shape[0], dtype=object)
         palette = _get_palette(cluster_ids.categories.size)
-        cluster_ids = cluster_ids[idx]
+        
         for k, cat in enumerate(cluster_ids.categories):
-            cell_colors[np.isin(cluster_ids, cat)] = palette[k]
+            cell_colors[cluster_ids == cat] = palette[k]
 
     if not switch_axes:
         cg = sns.clustermap(
             data=df,
             row_colors=cell_colors if not on_average else None,
             col_colors=None,
-            row_cluster=row_cluster,
-            col_cluster=col_cluster,
+            row_cluster=groupby_cluster,
+            col_cluster=attrs_cluster,
             linewidths=0,
             yticklabels=cluster_ids if on_average else [],
-            xticklabels=genes,
+            xticklabels=attr_names,
             figsize=panel_size,
             **kwargs,
         )
         cg.ax_heatmap.set_ylabel("")
+        if attrs_labelsize is not None:
+            cg.ax_heatmap.tick_params(axis='x', labelsize=attrs_labelsize)
     else:
         cg = sns.clustermap(
             data=df.T,
             row_colors=None,
             col_colors=cell_colors if not on_average else None,
-            row_cluster=row_cluster,
-            col_cluster=col_cluster,
+            row_cluster=attrs_cluster,
+            col_cluster=groupby_cluster,
             linewidths=0,
-            yticklabels=genes,
+            yticklabels=attr_names,
             xticklabels=cluster_ids if on_average else [],
             figsize=panel_size,
             **kwargs,
         )
         cg.ax_heatmap.set_xlabel("")
+        if attrs_labelsize is not None:
+            cg.ax_heatmap.tick_params(axis='y', labelsize=attrs_labelsize)
 
-    if row_cluster:
+    show_row_dendrogram = (attrs_cluster and attrs_dendrogram) if switch_axes else (groupby_cluster and groupby_dendrogram)
+    show_col_dendrogram = (groupby_cluster and groupby_dendrogram) if switch_axes else (attrs_cluster and attrs_dendrogram)
+
+    if show_row_dendrogram:
         cg.ax_heatmap.yaxis.tick_right()
-    else:
-        cg.ax_heatmap.yaxis.tick_left()
+        cg.ax_row_dendrogram.set_visible(True)
 
-    cg.ax_row_dendrogram.set_visible(row_cluster)
-    cg.cax.tick_params(labelsize=10)
-    cg.fig.dpi = dpi
-
-    if not row_cluster:
-        # Move the colorbar to the right-side.
-        color_box = cg.ax_heatmap.get_position()
-        color_box.x0 = color_box.x1 + 0.04
-        color_box.x1 = color_box.x0 + 0.02
-        cg.cax.set_position(color_box)
-        cg.cax.yaxis.set_ticks_position("right")
-    else:
         # Avoid overlap of colorbar and row dendrogram.
-        color_box = cg.cax.get_position()
+        color_box = cg.ax_cbar.get_position()
         square_plot = cg.ax_heatmap.get_position()
         if square_plot.y1 > color_box.y0:
             y_diff = square_plot.y1 - color_box.y0
             color_box.y0 = square_plot.y1
             color_box.y1 += y_diff
-            cg.cax.set_position(color_box)
+            cg.ax_cbar.set_position(color_box)
+    else:
+        cg.ax_heatmap.yaxis.tick_left()
+        cg.ax_row_dendrogram.set_visible(False)
+
+        # Move the colorbar to the right-side.
+        color_box = cg.ax_heatmap.get_position()
+        color_box.x0 = color_box.x1 + 0.04
+        color_box.x1 = color_box.x0 + 0.02
+        cg.ax_cbar.set_position(color_box)
+        cg.ax_cbar.yaxis.set_ticks_position("right")
+
+
+    if show_col_dendrogram:
+        cg.ax_heatmap.xaxis.tick_bottom()
+        cg.ax_col_dendrogram.set_visible(True)
+    else:
+        cg.ax_heatmap.xaxis.tick_top()
+        cg.ax_col_dendrogram.set_visible(False)
+    
+    cg.ax_cbar.tick_params(labelsize=cbar_labelsize)
+    cg.fig.dpi = dpi
 
     if not on_average:
-        orientation = 'left' if not switch_axes else 'top'
-        _plot_cluster_labels_in_heatmap(cg.ax_heatmap, cluster_ids, orientation)
+        if groupby_cluster:
+            from matplotlib.patches import Patch
+            legend_elements = [Patch(color = color, label = label) for color, label in zip(palette, cluster_ids.categories)]
+            cg.ax_heatmap.legend(handles=legend_elements, loc='lower left', bbox_to_anchor = (1.02, 1.02), fontsize = groupby_labelsize)           
+        else:
+            values = cluster_ids.value_counts().values
+            ticks = np.cumsum(values) - values / 2
+            labels = cluster_ids.categories
+            if not switch_axes:
+                cg.ax_row_colors.yaxis.tick_left()
+                cg.ax_row_colors.set_yticks(ticks)
+                cg.ax_row_colors.set_yticklabels(labels)
+                cg.ax_row_colors.tick_params(axis='y', left = False, length=10)
+            else:
+                cg.ax_col_colors.xaxis.tick_top()
+                cg.ax_col_colors.set_xticks(ticks)
+                cg.ax_col_colors.set_xticklabels(labels, rotation=45)
+                cg.ax_col_colors.tick_params(axis='x', top = False, labelsize = groupby_labelsize, length=10)
 
     if not isinstance(data, anndata.AnnData):
         if cur_matkey != data.current_matrix():
@@ -969,9 +1038,7 @@ def dotplot(
     keywords.update(kwds)
 
     from scipy.sparse import issparse
-    X = data[:, genes].X
-    if issparse(X):
-        X = X.toarray()
+    X = slicing(data[:, genes].X)
     df = pd.DataFrame(data=X, columns=genes)
     df[groupby] = data.obs[groupby].values
 
@@ -981,7 +1048,7 @@ def dotplot(
     summarized_df = df.groupby(groupby).aggregate([reduce_function, non_zero])
     if sort_function is not None:
         row_indices = sort_function(summarized_df)
-        summarized_df = summarized_df.iloc[row_indices]
+        summarized_df = summarized_df.loc[row_indices]
     else:
         summarized_df = summarized_df.loc[natsorted(summarized_df.index, reverse=True)]
 
@@ -1191,8 +1258,7 @@ def dendrogram(
         embed_df = pd.DataFrame(X_from_rep(data, rep))
         embed_df.set_index(data.obs[groupby], inplace=True)
     else:
-        sub_data = data[:, genes]
-        X = sub_data.X.toarray() if issparse(sub_data.X) else sub_data.X
+        X = slicing(data[:, genes].X)
         embed_df = pd.DataFrame(X)
         embed_df.set_index(data.obs[groupby], inplace=True)
 
@@ -1282,7 +1348,7 @@ def hvfplot(
 
     ax.scatter(x[hvg_index], y[hvg_index], s=5, c='b', marker='o', linewidth=0.5, alpha=0.5, label='highly variable features')
     ax.scatter(x[~hvg_index], y[~hvg_index], s=5, c='k', marker='o', linewidth=0.5, alpha=0.5, label = 'other features')
-    ax.legend(loc = 'upper right', fontsize = 5)
+    ax.legend(loc = 'best', fontsize = 5)
     ax.set_xlabel("Mean log expression")
     ax.set_ylabel("Variance of log expression")
 
@@ -1629,8 +1695,10 @@ def rank_plot(
 def ridgeplot(
     data: Union[MultimodalData, UnimodalData],
     features: Union[str, List[str]],
+    donor_attr: Optional[str] = None,
     qc_attr: Optional[str] = None,
     overlap: Optional[float] = 0.5,
+    left_adjust: Optional[float] = 0.35,
     panel_size: Optional[Tuple[float, float]] = (6, 4),
     return_fig: Optional[bool] = False,
     dpi: Optional[float] = 300.0,
@@ -1645,10 +1713,14 @@ def ridgeplot(
         CITE-Seq or Cyto data.
     features : `str` or `List[str]`
         One or more features to display.
+    donor_attr: `str`, optional, default None
+        If not None, `features` must contain only one feature, plot this feature by donor indicated as `donor_attr`.
     qc_attr: `str`, optional, default None
         If not None, only data.obs[qc_attr] == True are used.
     overlap: `float`, default 0.5
         Overlap between adjacent ridge plots (top and bottom).
+    left_adjust: `float`, default 0.35
+        Left margin for displaying labels.
     panel_size: `tuple`, optional (default: `(6, 4)`)
         The plot size (width, height) in inches.
     return_fig: ``bool``, optional, default: ``False``
@@ -1665,6 +1737,7 @@ def ridgeplot(
     Examples
     --------
     >>> fig = pg.ridgeplot(data, features = ['CD8', 'CD4', 'CD3'], show = False, dpi = 500)
+    >>> fig = pg.ridgeplot(data, features = 'CD3', donor_attr = 'assignment', show = False, dpi = 500)
     """
     sns.set(style="white", rc={"axes.facecolor": (0, 0, 0, 0)})
 
@@ -1673,23 +1746,45 @@ def ridgeplot(
     if isinstance(features, str):
         features = [features]
     if len(features) > 8:
-        logger.warning("With with_control == False, only up to 8 features are allowed!")
+        logger.warning("At most 8 features are allowed to be plotted together!")
         return None
 
-    exprs = []
-    feats = []
+    df = None
+    if donor_attr is None:
+        exprs = []
+        feats = []
 
-    size = idx.sum()
-    for feature in features:
-        fid = data.var_names.get_loc(feature)
-        exprs.append(data.get_matrix("arcsinh.transformed")[idx, fid].toarray()[:, 0])
-        feats.append(np.repeat(feature, size))
+        size = idx.sum()
+        for feature in features:
+            fid = data.var_names.get_loc(feature)
+            exprs.append(slicing(data.get_matrix("arcsinh.transformed"), idx, fid))
+            feats.append(np.repeat(feature, size))
 
-    df = pd.DataFrame({"expression": np.concatenate(exprs), "feature": np.concatenate(feats)})
+        df = pd.DataFrame({"expression": np.concatenate(exprs), "feature": np.concatenate(feats)})
+    else:
+        if len(features) != 1:
+            logger.warning("When donor_attr is set, only one feature can be provided!")
+            return None
+        if donor_attr not in data.obs:
+            logger.warning(f"{donor_attr} is not in data.obs!")
+            return None
+        feature = features[0]
+        if feature not in data.var_names:
+            logger.warning(f"Feature {feature} is not included in data.var_names!")
+            return None
+        fid = data.var_names.get_loc(features[0])
+        df = pd.DataFrame({"expression": slicing(data.get_matrix("arcsinh.transformed"), idx, fid), "feature": data.obs.loc[idx, donor_attr]})
 
     g = sns.FacetGrid(df, row="feature", hue="feature", aspect=8, height=1.0)
-    g.map(sns.kdeplot, "expression", clip_on=False, shade=True, alpha=1, lw=1.5)
-    g.map(sns.kdeplot, "expression", clip_on=False, color="k", lw=1)
+    try:
+        g.map(sns.kdeplot, "expression", clip_on=False, shade=True, alpha=1, lw=1.5)
+        g.map(sns.kdeplot, "expression", clip_on=False, color="k", lw=1)
+    except RuntimeError as re:
+        if str(re).startswith("Selected KDE bandwidth is 0. Cannot estimate density."):
+            g.map(sns.kdeplot, "expression", clip_on=False, shade=True, alpha=1, lw=1.5, bw=0.1)
+            g.map(sns.kdeplot, "expression", clip_on=False, color="k", lw=1, bw=0.1)
+        else:
+            raise re
     g.map(plt.axhline, y=0, lw=1, clip_on=False)
 
     def _set_label(value, color, label):
@@ -1698,13 +1793,15 @@ def ridgeplot(
 
     g.map(_set_label, "expression")
 
-    g.fig.subplots_adjust(hspace=-overlap)
+    g.fig.subplots_adjust(hspace=-overlap, left=left_adjust)
 
     g.set_titles("")
     g.set_xlabels("")
     g.set(yticks=[])
     g.despine(bottom=True, left=True)
 
+    if donor_attr is not None:
+        g.fig.suptitle(features[0], x = 0.0, y = 0.98, ha = "left")
     g.fig.set_dpi(dpi)
     g.fig.set_figwidth(panel_size[0])
     g.fig.set_figheight(panel_size[1])
@@ -1712,130 +1809,3 @@ def ridgeplot(
     sns.reset_orig()
 
     return g.fig if return_fig else None
-
-
-
-def doublet_plot(
-    scores: List[float],
-    codes: List[int],
-    code_names: Optional[List[str]] = ["singlet", "singlet2", "doublet"],
-    panel_size: Optional[Tuple[float, float]] = (4, 3),
-    left: Optional[float] = 0.2,
-    bottom: Optional[float] = 0.2,
-    wspace: Optional[float] = 0.2,
-    hspace: Optional[float] = 0.2,
-    return_fig: Optional[bool] = False,
-    dpi: Optional[float] = 300.0,
-    **kwargs,
-) -> Union[plt.Figure, None]:
-    """Generate KDE for doublet predictions
-
-    Parameters
-    ----------
-    scores: `List[float]`,
-        Doublet scores.
-    codes: `List[int]`
-        Doublet type codes.
-    code_names: `List[str]`, optional (default: `["singlet", "singlet2", "doublet"]`)
-        Doublet type names.
-    panel_size: `tuple`, optional (default: `(4, 3)`)
-        The plot size (width, height) in inches.
-    left: `float`, optional (default: `0.15`)
-        This parameter sets the figure's left margin as a fraction of panel's width (left * panel_size[0]).
-    bottom: `float`, optional (default: `0.15`)
-        This parameter sets the figure's bottom margin as a fraction of panel's height (bottom * panel_size[1]).
-    wspace: `float`, optional (default: `0.3`)
-        This parameter sets the width between panels and also the figure's right margin as a fraction of panel's width (wspace * panel_size[0]).
-    hspace: `float`, optional (defualt: `0.15`)
-        This parameter sets the height between panels and also the figure's top margin as a fraction of panel's height (hspace * panel_size[1]).
-    return_fig: ``bool``, optional, default: ``False``
-        Return a ``Figure`` object if ``True``; return ``None`` otherwise.
-    dpi: ``float``, optional, default: ``300.0``
-        The resolution in dots per inch.
-
-    Returns
-    -------
-
-    `Figure` object
-        A ``matplotlib.figure.Figure`` object containing the dot plot if ``return_fig == True``
-
-    Examples
-    --------
-    """
-    from scipy.stats import gaussian_kde
-
-    # constants
-    alpha = 0.7
-    eps = 1e-6
-
-    fig, axes = _get_subplot_layouts(nrows = 2, ncols = 2, panel_size=panel_size, dpi=dpi, left=left, bottom=bottom, wspace=wspace, hspace=hspace, sharex=False, sharey=False)
-
-    # Panel 1, KDE of scores
-    ax = axes[0, 0]
-    x = np.linspace(scores.min() - 0.05, scores.max() + 0.05, 200) # generate x coordinates
-    kde = gaussian_kde(scores)
-    y = kde(x)
-    ax.plot(x, y, '-', c='k')
-    ax.set_ylim(bottom = 0.0)
-    ax.set_xlabel('doublet score')
-    ax.set_ylabel('density')
-
-    # Panel 2, KDE of log-transformed scores
-    ax = axes[0, 1]
-    scores = np.log(scores)
-    x = np.linspace(scores.min() - 0.5, scores.max() + 0.5, 200) # generate x coordinates
-    kde = gaussian_kde(scores)
-    y = kde(x)
-    ax.plot(x, y, '-', c='k')
-    ax.set_ylim(bottom = 0.0)
-    ax.set_xlabel('log doublet score')
-    ax.set_ylabel('density')
-
-    # Panel 3, KDE of each cluster
-    ax = axes[1, 0]
-    # Partition by clusters
-    idx1 = codes == 0
-    idx2 = codes == 1
-    idx3 = codes == 2
-    # Get scale factor for each cluster
-    scale1 = idx1.sum() * 1.0 / codes.size
-    scale2 = idx2.sum() * 1.0 / codes.size
-    scale3 = idx3.sum() * 1.0 / codes.size
-    # kde for singlets1
-    kde1 = gaussian_kde(scores[idx1])
-    y1 = kde1(x) * scale1
-    idxs1 = y1 > eps
-    ax.plot(x[idxs1], y1[idxs1], '-', c = 'orange', label = code_names[0])
-    # kde for singlets2
-    kde2 = gaussian_kde(scores[idx2])
-    y2 = kde2(x) * scale2
-    idxs2 = y2 > eps
-    ax.plot(x[idxs2], y2[idxs2], '-', c = 'green', label = code_names[1])
-    # kde for doublets
-    if idx3.sum() > 0:
-        scores3 = scores[idx3]
-        if np.std(scores3) < eps: # if too small, add jitters
-            scores3 += np.random.normal(scale = 0.01, size = scores3.size)
-        kde3 = gaussian_kde(scores3)
-        y3 = kde3(x) * scale3
-        idxs3 = y3 > eps
-        ax.plot(x[idxs3], y3[idxs3], '-', c = 'red', label = code_names[2])
-    # Set ylim bottom to 0 and show legend
-    ax.set_ylim(bottom = 0)
-    ax.set_xlabel('log doublet score')
-    ax.set_ylabel('density')
-    ax.legend(loc = 'best')
-
-    # Panel 4, overlay the overall KDE with cluster-specific KDEs
-    ax = axes[1, 1]
-    ax.plot(x, y, '-', c='k', label = 'overall')
-    ax.plot(x[idxs1], y1[idxs1], '-', c = 'orange', alpha = alpha, label = 'singlet1')
-    ax.plot(x[idxs2], y2[idxs2], '-', c = 'green', alpha = alpha, label = 'singlet2')
-    if idx3.sum() > 0:
-        ax.plot(x[idxs3], y3[idxs3], '-', c = 'red', alpha = alpha, label = 'doublet')
-    ax.set_ylim(bottom = 0)
-    ax.set_xlabel('log doublet score')
-    ax.set_ylabel('density')
-    ax.legend(loc = 'best')
-
-    return fig if return_fig else None
