@@ -6,11 +6,9 @@ from typing import List, Union, Tuple
 
 
 def _transform_basis(basis: str) -> str:
-    if basis == "tsne" or basis == "citeseq_tsne":
+    if basis == "tsne":
         return "tSNE"
-    elif basis == "fitsne":
-        return "FItSNE"
-    elif basis == "umap":
+    elif basis == "umap" or basis == "citeseq_umap":
         return "UMAP"
     elif basis == "diffmap":
         return "DC"
@@ -20,10 +18,6 @@ def _transform_basis(basis: str) -> str:
         return "DPC"
     elif basis == "fle":
         return "FLE"
-    elif basis == "net_tsne":
-        return "Net-tSNE"
-    elif basis == "net_fitsne":
-        return "Net-FItSNE"
     elif basis == "net_umap":
         return "Net-UMAP"
     elif basis == "net_fle":
@@ -294,36 +288,6 @@ def _get_palette(n_labels: int, with_background: bool = False, show_background:
 
     return palette
 
-def _plot_cluster_labels_in_heatmap(ax, ids, orientation):
-    pos_acc = 0
-    ticks = []
-    labels = []
-    max_label_len = 0
-    for label, value in ids.value_counts().iteritems():
-        labels.append(label)
-        if len(label) > max_label_len:
-            max_label_len = len(label)
-        ticks.append(pos_acc + value / 2)
-        pos_acc += value
-
-    ax.grid(False)
-
-    if orientation == 'left':
-        ax.set_yticks(ticks)
-        ax.set_yticklabels(labels, rotation=0, x=-0.05)
-        ax.tick_params(axis='y', left=False)
-    else:
-        color_box = ax.get_position()
-        ax.set_xticks(ticks)
-        ax.set_xticklabels(labels, rotation=90, y=color_box.y1+max_label_len/100)
-        ax.xaxis.set_ticks_position('top')
-        ax.tick_params(axis='x', top=False)
-
-    # Remove borders
-    ax.spines['left'].set_visible(False)
-    ax.spines['right'].set_visible(False)
-    ax.spines['top'].set_visible(False)
-    ax.spines['bottom'].set_visible(False)
 
 
 Restriction = namedtuple("Restriction", ["negation", "values"])
