@@ -2,12 +2,13 @@ import time
 import numpy as np
 import pandas as pd
 
-from scipy.sparse import issparse
+from scipy.sparse import issparse, csr_matrix
 
-from sklearn.decomposition import PCA
+from sklearn.decomposition import PCA, TruncatedSVD
 
 from typing import List, Tuple, Union
 from pegasusio import UnimodalData, MultimodalData, calc_qc_filters, DictWithDefault
+from pegasus.tools import normalize_by_count
 
 import logging
 logger = logging.getLogger(__name__)
@@ -146,6 +147,7 @@ def get_filter_stats(data: MultimodalData, min_genes_before_filt: int = 100) ->
     return df_cells
 
 
+@timer(logger=logger)
 def filter_data(data: MultimodalData, focus_list: List[str] = None) -> None:
     """ Filter data based on qc_metrics calculated in ``pg.qc_metrics``.
 
@@ -169,6 +171,7 @@ def filter_data(data: MultimodalData, focus_list: List[str] = None) -> None:
     data.filter_data(focus_list = focus_list, cache_passqc = True)
 
 
+@timer(logger=logger)
 def identify_robust_genes(data: MultimodalData, percent_cells: float = 0.05) -> None:
     """ Identify robust genes as candidates for HVG selection and remove genes that are not expressed in any cells.
 
@@ -197,13 +200,17 @@ def identify_robust_genes(data: MultimodalData, percent_cells: float = 0.05) ->
     if isinstance(data, MultimodalData):
         data = data.current_data()
 
-    data.var["n_cells"] = data.X.getnnz(axis=0)
-    data.var["percent_cells"] = (data.var["n_cells"] / data.shape[0]) * 100
-    data.var["robust"] = data.var["percent_cells"] >= percent_cells
-    data.var["highly_variable_features"] = data.var["robust"]  # default all robust genes are "highly" variable
-
     prior_n = data.shape[1]
-    data._inplace_subset_var(data.var["n_cells"] > 0)
+
+    if issparse(data.X):
+        data.var["n_cells"] = data.X.getnnz(axis=0)
+        data._inplace_subset_var(data.var["n_cells"] > 0)
+        data.var["percent_cells"] = (data.var["n_cells"] / data.shape[0]) * 100
+        data.var["robust"] = data.var["percent_cells"] >= percent_cells
+    else:
+        data.var["robust"] = True
+
+    data.var["highly_variable_features"] = data.var["robust"]  # default all robust genes are "highly" variable
     logger.info(f"After filtration, {data.shape[1]}/{prior_n} genes are kept. Among {data.shape[1]} genes, {data.var['robust'].sum()} genes are robust.")
 
 
@@ -294,8 +301,11 @@ def _run_filter_data(
 
 
 @timer(logger=logger)
-@run_gc
-def log_norm(data: MultimodalData, norm_count: float = 1e5, backup_matrix: str = "raw.X") -> None:
+def log_norm(
+    data: MultimodalData, 
+    norm_count: float = 1e5, 
+    backup_matrix: str = "raw.X",
+) -> None:
     """Normalization, and then apply natural logarithm to the data.
 
     Parameters
@@ -307,7 +317,7 @@ def log_norm(data: MultimodalData, norm_count: float = 1e5, backup_matrix: str =
         Total count of cells after normalization.
 
     backup_matrix: ``str``, optional, default: ``raw.X``.
-        Where to back up the count matrix.
+        The key name of the backup count matrix, usually the raw counts.
 
     Returns
     -------
@@ -315,6 +325,8 @@ def log_norm(data: MultimodalData, norm_count: float = 1e5, backup_matrix: str =
 
     Update ``data.X`` with count matrix after log-normalization. In addition, back up the original count matrix as ``backup_matrix``.
 
+    In case of rerunning normalization while ``backup_matrix`` already exists, use ``backup_matrix`` instead of ``data.X`` for normalization.
+
     Examples
     --------
     >>> pg.log_norm(data)
@@ -324,19 +336,27 @@ def log_norm(data: MultimodalData, norm_count: float = 1e5, backup_matrix: str =
 
     assert data.get_modality() == "rna"
 
-    data.add_matrix(backup_matrix, data.X)
-    data.X = data.get_matrix("X").astype(np.float32)
+    if backup_matrix not in data.list_keys():
+        data.add_matrix(backup_matrix, data.X)
+        data.X = data.X.astype(np.float32)  # force copy
+    else:
+        # The case of rerunning log_norm. Use backup matrix as source.
+        data.X = data.get_matrix(backup_matrix).astype(np.float32)  # force copy
+        logger.warning("Rerun log-normalization. Use the raw counts in backup instead.")
 
-    mat = data.X[:, data.var["robust"].values]
-    scale = norm_count / mat.sum(axis=1).A1
-    data.X.data *= np.repeat(scale, np.diff(data.X.indptr))
-    data.X.data = np.log1p(data.X.data) # faster than data.X.log1p()
-    data.obs["scale"] = scale
+    data.obs["scale"] = normalize_by_count(data.X, data.var["robust"].values, norm_count, True)
+    data.uns["norm_count"] = norm_count
 
 
 @run_gc
-def select_features(data: MultimodalData, features: str = "highly_variable_features", standardize: bool = True, max_value: float = 10.0) -> str:
-    """ Subset the features and store the resulting matrix in dense format in data.uns with `'fmat_'` prefix, with the option of standardization and truncating based on max_value. `'fmat_*'` will be removed before writing out the disk.
+def select_features(
+    data: MultimodalData,
+    features: str = "highly_variable_features",
+    standardize: bool = True,
+    max_value: float = 10.0,
+    use_cache: bool = True,
+) -> str:
+    """ Subset the features and store the resulting matrix in dense format in data.uns with `'_tmp_fmat_'` prefix, with the option of standardization and truncating based on max_value. `'_tmp_fmat_*'` will be removed before writing out the disk.
 
     Parameters
     ----------
@@ -352,12 +372,15 @@ def select_features(data: MultimodalData, features: str = "highly_variable_featu
     max_value: ``float``, optional, default: ``10``.
         The threshold to truncate data after scaling. If ``None``, do not truncate.
 
+    use_cache: ``bool``, optional, default: ``False``.
+        Whether to use precalculated subset matrix if exists.
+
     Returns
     -------
     keyword: ``str``
         The keyword in ``data.uns`` referring to the features selected.
 
-    Update ``data.uns``:
+    Update ``data.uns`` if needed:
 
         * ``data.uns[keyword]``: A submatrix of the data containing features selected.
 
@@ -365,14 +388,17 @@ def select_features(data: MultimodalData, features: str = "highly_variable_featu
     --------
     >>> pg.select_features(data)
     """
-    keyword = "fmat_" + str(features)  # fmat: feature matrix
+    keyword = "_tmp_fmat_" + str(features)  # fmat: feature matrix
 
-    if features is not None:
-        assert features in data.var
-        X = data.X[:, data.var[features].values]
-    else:
-        X = data.X
-    data.uns[keyword] = X.toarray() if issparse(X) else X.copy()
+    if (not use_cache) or (use_cache and (keyword not in data.uns)):
+        if features is not None:
+            assert features in data.var
+            X = data.X[:, data.var[features].values]
+        else:
+            X = data.X
+
+        from pegasus.tools import slicing
+        data.uns[keyword] = slicing(X, copy=True)
 
     if standardize or (max_value is not None):
         X = data.uns[keyword]
@@ -383,7 +409,10 @@ def select_features(data: MultimodalData, features: str = "highly_variable_featu
             std[std == 0] = 1
             X -= m1
             X /= std
+            data.uns["stdzn_mean"] = m1
+            data.uns["stdzn_std"] = std
         if max_value is not None:
+            data.uns["stdzn_max_value"] = max_value
             X[X > max_value] = max_value
             X[X < -max_value] = -max_value
         data.uns[keyword] = X
@@ -397,9 +426,10 @@ def pca(
     n_components: int = 50,
     features: str = "highly_variable_features",
     standardize: bool = True,
-    max_value: float = 10,
+    max_value: float = 10.0,
     robust: bool = False,
     random_state: int = 0,
+    use_cache: bool = True,
 ) -> None:
     """Perform Principle Component Analysis (PCA) to the data.
 
@@ -423,11 +453,13 @@ def pca(
         The threshold to truncate data after scaling. If ``None``, do not truncate.
 
     robust: ``bool``, optional, default: ``False``.
-        If true, use 'arpack' instead of 'randomized' for large sparse matrices (i.e. max(X.shape) > 500 and n_components < 0.8 * min(X.shape))
+        If true, use 'arpack' instead of 'randomized' for large matrices (i.e. max(X.shape) > 500 and n_components < 0.8 * min(X.shape))
 
     random_state: ``int``, optional, default: ``0``.
         Random seed to be set for reproducing result.
 
+    use_cache: ``bool``, optional, default: ``True``.
+        Whether to use precalculated subset feature matrix if exists.
 
     Returns
     -------
@@ -449,32 +481,84 @@ def pca(
     --------
     >>> pg.pca(data)
     """
-    keyword = select_features(data, features = features, standardize = standardize, max_value = max_value)
+    keyword = select_features(data, features=features, standardize=standardize, max_value=max_value, use_cache=use_cache)
     X = data.uns[keyword]
 
-    pca = PCA(n_components=n_components, random_state=random_state)
     if robust:
         svd_solver = "arpack" if max(X.shape) > 500 and n_components < 0.8 * min(X.shape) else "full"
         pca = PCA(n_components=n_components, random_state=random_state, svd_solver=svd_solver)
+    else:
+        pca = PCA(n_components=n_components, random_state=random_state)
 
     X_pca = pca.fit_transform(X)
 
     data.obsm["X_pca"] = np.ascontiguousarray(X_pca)
-    data.uns[
-        "PCs"
-    ] = pca.components_.T  # cannot be varm because numbers of features are not the same
+    data.uns["PCs"] = np.ascontiguousarray(pca.components_.T)  # cannot be varm because numbers of features are not the same
     data.uns["pca"] = {}
     data.uns["pca"]["variance"] = pca.explained_variance_
     data.uns["pca"]["variance_ratio"] = pca.explained_variance_ratio_
+    data.uns["pca_features"] = features # record which feature to use
 
 
 @timer(logger=logger)
-def pc_regress_out(
+def pc_transform(
+    data: MultimodalData,
+    X: Union[csr_matrix, np.ndarray]
+) -> np.ndarray:
+    """Transform a new count matrix X using existing PCs contained in data.
+
+    Parameters
+    ----------
+    data: ``pegasusio.MultimodalData``
+        Annotated data matrix with PCs calculated
+
+    X: ``Union[csr_matrix, np.ndarray]``
+        New input count matrix to be transformed. Number of genes must be the same as in data
+
+    Returns
+    -------
+    Transformed PC coordinates for X.
+
+    Examples
+    --------
+    >>> pg.pc_transform(data, X_new)
+    """
+    assert data.shape[1] == X.shape[1]
+    if ("norm_count" not in data.uns) or ("robust" not in data.var) or ("pca_features" not in data.uns) or ("PCs" not in data.uns):
+        raise ValueError("pc_transform cannot be applied before identify_robust_gene, highly_variable_features and pca are called!")
+
+    # normalize and log transform X
+    X = X.astype(np.float32) # convert to float
+    _ = normalize_by_count(X, data.var["robust"].values, data.uns["norm_count"], True)
+
+    # select features
+    if data.uns["pca_features"] is not None:
+        X = X[:, data.var[data.uns["pca_features"]].values]
+    from pegasus.tools import slicing
+    X = slicing(X, copy=True)
+
+    # standardization
+    if "stdzn_mean" in data.uns:
+        assert "stdzn_std" in data.uns
+        X -= data.uns["stdzn_mean"]
+        X /= data.uns["stdzn_std"]
+    if "stdzn_max_value" in data.uns:
+        max_value = data.uns["stdzn_max_value"]
+        X[X > max_value] = max_value
+        X[X < -max_value] = -max_value
+
+    # transform to PC coordinates
+    X_pca = np.matmul(X, data.uns["PCs"])
+    return X_pca
+
+
+@timer(logger=logger)
+def regress_out(
     data: Union[MultimodalData, UnimodalData],
     attrs: List[str],
     rep: str = 'pca',
 ) -> str:
-    """Regress out effects due to specific observational attributes at Principal Component level.
+    """Regress out effects due to specific observational attributes.
 
     Parameters
     ----------
@@ -490,7 +574,7 @@ def pc_regress_out(
 
     Returns
     -------
-    pca_key: ``str``
+    res_key: ``str``
         The key to the resulting new embedding matrix in ``data.obsm``. It's ``'X_'+rep+'_regressed'``.
 
     Update ``data.obsm``:
@@ -498,7 +582,7 @@ def pc_regress_out(
 
     Examples
     --------
-    >>> pg.pc_regress(data, attrs=['G1/S', 'G2/M'])
+    >>> pg.regress_out(data, attrs=['G1/S', 'G2/M'])
     """
     n_components = data.obsm[f'X_{rep}'].shape[1]
 
@@ -523,3 +607,104 @@ def pc_regress_out(
     data.obsm[f'X_{pca_key}'] = np.vstack(response_list).T.astype(data.obsm[f'X_{rep}'].dtype)
 
     return pca_key
+
+
+@timer(logger=logger)
+def tsvd(
+    data: MultimodalData,
+    n_components: int = 51,
+    features: str = "robust",
+    robust: bool = False,
+    random_state: int = 0,
+) -> None:
+    """Perform Truncated Singular Value Decomposition (TSVD) to the data.
+
+    The calculation uses *scikit-learn* implementation.
+
+    Parameters
+    ----------
+    data: ``pegasusio.MultimodalData``
+        Annotated data matrix with rows for cells and columns for genes.
+
+    n_components: ``int``, optional, default: ``51``.
+        Number of components to compute. Because we do not remove means, the first component might be related to the mean values and thus we ask for one more component than PCA as default (i.e. 51).
+
+    features: ``str``, optional, default: ``"robust"``.
+        Keyword in ``data.var`` to specify features used for TSVD.
+
+    robust: ``bool``, optional, default: ``False``.
+        If true, use 'arpack' instead of 'randomized'.
+
+    random_state: ``int``, optional, default: ``0``.
+        Random seed to be set for reproducing result.
+
+    Returns
+    -------
+    ``None``.
+
+    Update ``data.obsm``:
+
+        * ``data.obsm["X_tsvd"]``: TSVD coordinates of the data.
+
+    Update ``data.uns``:
+
+        * ``data.uns["TSVDs"]``: The TSVD components containing the loadings.
+
+    Examples
+    --------
+    >>> pg.tsvd(data)
+    """
+    X = data.X
+    if (features is not None) and (data.var[features].sum() < data.shape[1]):
+        X = X[:, data.var[features].values]
+
+    tsvd = TruncatedSVD(n_components = n_components, algorithm = 'arpack' if robust else 'randomized', random_state = random_state)
+    X_tsvd = tsvd.fit_transform(X)
+
+    data.obsm["X_tsvd"] = np.ascontiguousarray(X_tsvd)
+    data.uns["TSVDs"] = np.ascontiguousarray(tsvd.components_.T)  # cannot be varm because numbers of features are not the same
+    data.uns["tsvd"] = {}
+    data.uns["tsvd"]["variance"] = tsvd.explained_variance_
+    data.uns["tsvd"]["variance_ratio"] = tsvd.explained_variance_ratio_
+    data.uns["tsvd_features"] = features # record which feature to use
+
+
+@timer(logger=logger)
+def tsvd_transform(
+    data: MultimodalData,
+    X: Union[csr_matrix, np.ndarray]
+) -> np.ndarray:
+    """Transform a new count matrix X using existing TSVD components contained in data.
+
+    Parameters
+    ----------
+    data: ``pegasusio.MultimodalData``
+        Annotated data matrix with PCs calculated
+
+    X: ``Union[csr_matrix, np.ndarray]``
+        New input count matrix to be transformed. Number of genes must be the same as in data
+
+    Returns
+    -------
+    Transformed TSVD coordinates for X.
+
+    Examples
+    --------
+    >>> pg.tsvd_transform(data, X_new)
+    """
+    assert data.shape[1] == X.shape[1]
+    if ("norm_count" not in data.uns) or ("robust" not in data.var) or ("tsvd_features" not in data.uns) or ("TSVDs" not in data.uns):
+        raise ValueError("tsvd_transform cannot be applied before identify_robust_gene, and tsvd are called!")
+
+    # normalize and log transform X
+    X = X.astype(np.float32) # convert to float
+    _ = normalize_by_count(X, data.var["robust"].values, data.uns["norm_count"], True)
+
+    # select features
+    if (data.uns["tsvd_features"] is not None) and (data.var[data.uns["tsvd_features"]].sum() < X.shape[1]):
+        X = X[:, data.var[data.uns["tsvd_features"]].values]
+
+    # transform to TSVD coordinates
+    from sklearn.utils.extmath import safe_sparse_dot
+    X_tsvd = safe_sparse_dot(X, data.uns["TSVDs"])
+    return X_tsvd
